#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: model_center_service.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import service as _service
from google.protobuf import service_reflection
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import message_pb2 as message__pb2
from google.api import annotations_pb2 as google_dot_api_dot_annotations__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='model_center_service.proto',
  package='ai_flow',
  syntax='proto3',
  serialized_options=b'\n\020com.aiflow.protoZ\010/ai_flow\210\001\001\220\001\001',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x1amodel_center_service.proto\x12\x07\x61i_flow\x1a\rmessage.proto\x1a\x1cgoogle/api/annotations.proto\"W\n\x1c\x43reateRegisteredModelRequest\x12\x37\n\x10registered_model\x18\x01 \x01(\x0b\x32\x1d.ai_flow.RegisteredModelParam\"\x84\x01\n\x1cUpdateRegisteredModelRequest\x12+\n\nmodel_meta\x18\x01 \x01(\x0b\x32\x17.ai_flow.ModelMetaParam\x12\x37\n\x10registered_model\x18\x02 \x01(\x0b\x32\x1d.ai_flow.RegisteredModelParam\"K\n\x1c\x44\x65leteRegisteredModelRequest\x12+\n\nmodel_meta\x18\x01 \x01(\x0b\x32\x17.ai_flow.ModelMetaParam\"\x1d\n\x1bListRegisteredModelsRequest\"N\n\x1fGetRegisteredModelDetailRequest\x12+\n\nmodel_meta\x18\x01 \x01(\x0b\x32\x17.ai_flow.ModelMetaParam\"{\n\x19\x43reateModelVersionRequest\x12+\n\nmodel_meta\x18\x01 \x01(\x0b\x32\x17.ai_flow.ModelMetaParam\x12\x31\n\rmodel_version\x18\x02 \x01(\x0b\x32\x1a.ai_flow.ModelVersionParam\"{\n\x19UpdateModelVersionRequest\x12+\n\nmodel_meta\x18\x01 \x01(\x0b\x32\x17.ai_flow.ModelMetaParam\x12\x31\n\rmodel_version\x18\x02 \x01(\x0b\x32\x1a.ai_flow.ModelVersionParam\"H\n\x19\x44\x65leteModelVersionRequest\x12+\n\nmodel_meta\x18\x01 \x01(\x0b\x32\x17.ai_flow.ModelMetaParam\"K\n\x1cGetModelVersionDetailRequest\x12+\n\nmodel_meta\x18\x01 \x01(\x0b\x32\x17.ai_flow.ModelMetaParam2\xd4\t\n\x12ModelCenterService\x12\x88\x01\n\x15\x63reateRegisteredModel\x12%.ai_flow.CreateRegisteredModelRequest\x1a\x11.ai_flow.Response\"5\x82\xd3\xe4\x93\x02/\"*/aiflow/model-center/register-model/create:\x01*\x12\x88\x01\n\x15updateRegisteredModel\x12%.ai_flow.UpdateRegisteredModelRequest\x1a\x11.ai_flow.Response\"5\x82\xd3\xe4\x93\x02/2*/aiflow/model-center/register-model/update:\x01*\x12\x88\x01\n\x15\x64\x65leteRegisteredModel\x12%.ai_flow.DeleteRegisteredModelRequest\x1a\x11.ai_flow.Response\"5\x82\xd3\xe4\x93\x02/2*/aiflow/model-center/register-model/update:\x01*\x12\x81\x01\n\x14listRegisteredModels\x12$.ai_flow.ListRegisteredModelsRequest\x1a\x11.ai_flow.Response\"0\x82\xd3\xe4\x93\x02*\x12(/aiflow/model-center/register-model/list\x12\x88\x01\n\x18getRegisteredModelDetail\x12(.ai_flow.GetRegisteredModelDetailRequest\x1a\x11.ai_flow.Response\"/\x82\xd3\xe4\x93\x02)\x12\'/aiflow/model-center/register-model/get\x12\x81\x01\n\x12\x63reateModelVersion\x12\".ai_flow.CreateModelVersionRequest\x1a\x11.ai_flow.Response\"4\x82\xd3\xe4\x93\x02.\")/aiflow/model-center/model-version/create:\x01*\x12\x81\x01\n\x12updateModelVersion\x12\".ai_flow.UpdateModelVersionRequest\x1a\x11.ai_flow.Response\"4\x82\xd3\xe4\x93\x02.2)/aiflow/model-center/model-version/update:\x01*\x12\x81\x01\n\x12\x64\x65leteModelVersion\x12\".ai_flow.DeleteModelVersionRequest\x1a\x11.ai_flow.Response\"4\x82\xd3\xe4\x93\x02.2)/aiflow/model-center/model-version/delete:\x01*\x12\x81\x01\n\x15getModelVersionDetail\x12%.ai_flow.GetModelVersionDetailRequest\x1a\x11.ai_flow.Response\".\x82\xd3\xe4\x93\x02(\x12&/aiflow/model-center/model-version/getB\"\n\x10\x63om.aiflow.protoZ\x08/ai_flow\x88\x01\x01\x90\x01\x01\x62\x06proto3'
  ,
  dependencies=[message__pb2.DESCRIPTOR,google_dot_api_dot_annotations__pb2.DESCRIPTOR,])




_CREATEREGISTEREDMODELREQUEST = _descriptor.Descriptor(
  name='CreateRegisteredModelRequest',
  full_name='ai_flow.CreateRegisteredModelRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='registered_model', full_name='ai_flow.CreateRegisteredModelRequest.registered_model', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=84,
  serialized_end=171,
)


_UPDATEREGISTEREDMODELREQUEST = _descriptor.Descriptor(
  name='UpdateRegisteredModelRequest',
  full_name='ai_flow.UpdateRegisteredModelRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='model_meta', full_name='ai_flow.UpdateRegisteredModelRequest.model_meta', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='registered_model', full_name='ai_flow.UpdateRegisteredModelRequest.registered_model', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=174,
  serialized_end=306,
)


_DELETEREGISTEREDMODELREQUEST = _descriptor.Descriptor(
  name='DeleteRegisteredModelRequest',
  full_name='ai_flow.DeleteRegisteredModelRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='model_meta', full_name='ai_flow.DeleteRegisteredModelRequest.model_meta', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=308,
  serialized_end=383,
)


_LISTREGISTEREDMODELSREQUEST = _descriptor.Descriptor(
  name='ListRegisteredModelsRequest',
  full_name='ai_flow.ListRegisteredModelsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=385,
  serialized_end=414,
)


_GETREGISTEREDMODELDETAILREQUEST = _descriptor.Descriptor(
  name='GetRegisteredModelDetailRequest',
  full_name='ai_flow.GetRegisteredModelDetailRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='model_meta', full_name='ai_flow.GetRegisteredModelDetailRequest.model_meta', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=416,
  serialized_end=494,
)


_CREATEMODELVERSIONREQUEST = _descriptor.Descriptor(
  name='CreateModelVersionRequest',
  full_name='ai_flow.CreateModelVersionRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='model_meta', full_name='ai_flow.CreateModelVersionRequest.model_meta', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='model_version', full_name='ai_flow.CreateModelVersionRequest.model_version', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=496,
  serialized_end=619,
)


_UPDATEMODELVERSIONREQUEST = _descriptor.Descriptor(
  name='UpdateModelVersionRequest',
  full_name='ai_flow.UpdateModelVersionRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='model_meta', full_name='ai_flow.UpdateModelVersionRequest.model_meta', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='model_version', full_name='ai_flow.UpdateModelVersionRequest.model_version', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=621,
  serialized_end=744,
)


_DELETEMODELVERSIONREQUEST = _descriptor.Descriptor(
  name='DeleteModelVersionRequest',
  full_name='ai_flow.DeleteModelVersionRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='model_meta', full_name='ai_flow.DeleteModelVersionRequest.model_meta', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=746,
  serialized_end=818,
)


_GETMODELVERSIONDETAILREQUEST = _descriptor.Descriptor(
  name='GetModelVersionDetailRequest',
  full_name='ai_flow.GetModelVersionDetailRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='model_meta', full_name='ai_flow.GetModelVersionDetailRequest.model_meta', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=820,
  serialized_end=895,
)

_CREATEREGISTEREDMODELREQUEST.fields_by_name['registered_model'].message_type = message__pb2._REGISTEREDMODELPARAM
_UPDATEREGISTEREDMODELREQUEST.fields_by_name['model_meta'].message_type = message__pb2._MODELMETAPARAM
_UPDATEREGISTEREDMODELREQUEST.fields_by_name['registered_model'].message_type = message__pb2._REGISTEREDMODELPARAM
_DELETEREGISTEREDMODELREQUEST.fields_by_name['model_meta'].message_type = message__pb2._MODELMETAPARAM
_GETREGISTEREDMODELDETAILREQUEST.fields_by_name['model_meta'].message_type = message__pb2._MODELMETAPARAM
_CREATEMODELVERSIONREQUEST.fields_by_name['model_meta'].message_type = message__pb2._MODELMETAPARAM
_CREATEMODELVERSIONREQUEST.fields_by_name['model_version'].message_type = message__pb2._MODELVERSIONPARAM
_UPDATEMODELVERSIONREQUEST.fields_by_name['model_meta'].message_type = message__pb2._MODELMETAPARAM
_UPDATEMODELVERSIONREQUEST.fields_by_name['model_version'].message_type = message__pb2._MODELVERSIONPARAM
_DELETEMODELVERSIONREQUEST.fields_by_name['model_meta'].message_type = message__pb2._MODELMETAPARAM
_GETMODELVERSIONDETAILREQUEST.fields_by_name['model_meta'].message_type = message__pb2._MODELMETAPARAM
DESCRIPTOR.message_types_by_name['CreateRegisteredModelRequest'] = _CREATEREGISTEREDMODELREQUEST
DESCRIPTOR.message_types_by_name['UpdateRegisteredModelRequest'] = _UPDATEREGISTEREDMODELREQUEST
DESCRIPTOR.message_types_by_name['DeleteRegisteredModelRequest'] = _DELETEREGISTEREDMODELREQUEST
DESCRIPTOR.message_types_by_name['ListRegisteredModelsRequest'] = _LISTREGISTEREDMODELSREQUEST
DESCRIPTOR.message_types_by_name['GetRegisteredModelDetailRequest'] = _GETREGISTEREDMODELDETAILREQUEST
DESCRIPTOR.message_types_by_name['CreateModelVersionRequest'] = _CREATEMODELVERSIONREQUEST
DESCRIPTOR.message_types_by_name['UpdateModelVersionRequest'] = _UPDATEMODELVERSIONREQUEST
DESCRIPTOR.message_types_by_name['DeleteModelVersionRequest'] = _DELETEMODELVERSIONREQUEST
DESCRIPTOR.message_types_by_name['GetModelVersionDetailRequest'] = _GETMODELVERSIONDETAILREQUEST
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

CreateRegisteredModelRequest = _reflection.GeneratedProtocolMessageType('CreateRegisteredModelRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATEREGISTEREDMODELREQUEST,
  '__module__' : 'model_center_service_pb2'
  # @@protoc_insertion_point(class_scope:ai_flow.CreateRegisteredModelRequest)
  })
_sym_db.RegisterMessage(CreateRegisteredModelRequest)

UpdateRegisteredModelRequest = _reflection.GeneratedProtocolMessageType('UpdateRegisteredModelRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPDATEREGISTEREDMODELREQUEST,
  '__module__' : 'model_center_service_pb2'
  # @@protoc_insertion_point(class_scope:ai_flow.UpdateRegisteredModelRequest)
  })
_sym_db.RegisterMessage(UpdateRegisteredModelRequest)

DeleteRegisteredModelRequest = _reflection.GeneratedProtocolMessageType('DeleteRegisteredModelRequest', (_message.Message,), {
  'DESCRIPTOR' : _DELETEREGISTEREDMODELREQUEST,
  '__module__' : 'model_center_service_pb2'
  # @@protoc_insertion_point(class_scope:ai_flow.DeleteRegisteredModelRequest)
  })
_sym_db.RegisterMessage(DeleteRegisteredModelRequest)

ListRegisteredModelsRequest = _reflection.GeneratedProtocolMessageType('ListRegisteredModelsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTREGISTEREDMODELSREQUEST,
  '__module__' : 'model_center_service_pb2'
  # @@protoc_insertion_point(class_scope:ai_flow.ListRegisteredModelsRequest)
  })
_sym_db.RegisterMessage(ListRegisteredModelsRequest)

GetRegisteredModelDetailRequest = _reflection.GeneratedProtocolMessageType('GetRegisteredModelDetailRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETREGISTEREDMODELDETAILREQUEST,
  '__module__' : 'model_center_service_pb2'
  # @@protoc_insertion_point(class_scope:ai_flow.GetRegisteredModelDetailRequest)
  })
_sym_db.RegisterMessage(GetRegisteredModelDetailRequest)

CreateModelVersionRequest = _reflection.GeneratedProtocolMessageType('CreateModelVersionRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATEMODELVERSIONREQUEST,
  '__module__' : 'model_center_service_pb2'
  # @@protoc_insertion_point(class_scope:ai_flow.CreateModelVersionRequest)
  })
_sym_db.RegisterMessage(CreateModelVersionRequest)

UpdateModelVersionRequest = _reflection.GeneratedProtocolMessageType('UpdateModelVersionRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPDATEMODELVERSIONREQUEST,
  '__module__' : 'model_center_service_pb2'
  # @@protoc_insertion_point(class_scope:ai_flow.UpdateModelVersionRequest)
  })
_sym_db.RegisterMessage(UpdateModelVersionRequest)

DeleteModelVersionRequest = _reflection.GeneratedProtocolMessageType('DeleteModelVersionRequest', (_message.Message,), {
  'DESCRIPTOR' : _DELETEMODELVERSIONREQUEST,
  '__module__' : 'model_center_service_pb2'
  # @@protoc_insertion_point(class_scope:ai_flow.DeleteModelVersionRequest)
  })
_sym_db.RegisterMessage(DeleteModelVersionRequest)

GetModelVersionDetailRequest = _reflection.GeneratedProtocolMessageType('GetModelVersionDetailRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETMODELVERSIONDETAILREQUEST,
  '__module__' : 'model_center_service_pb2'
  # @@protoc_insertion_point(class_scope:ai_flow.GetModelVersionDetailRequest)
  })
_sym_db.RegisterMessage(GetModelVersionDetailRequest)


DESCRIPTOR._options = None

_MODELCENTERSERVICE = _descriptor.ServiceDescriptor(
  name='ModelCenterService',
  full_name='ai_flow.ModelCenterService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=898,
  serialized_end=2134,
  methods=[
  _descriptor.MethodDescriptor(
    name='createRegisteredModel',
    full_name='ai_flow.ModelCenterService.createRegisteredModel',
    index=0,
    containing_service=None,
    input_type=_CREATEREGISTEREDMODELREQUEST,
    output_type=message__pb2._RESPONSE,
    serialized_options=b'\202\323\344\223\002/\"*/aiflow/model-center/register-model/create:\001*',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='updateRegisteredModel',
    full_name='ai_flow.ModelCenterService.updateRegisteredModel',
    index=1,
    containing_service=None,
    input_type=_UPDATEREGISTEREDMODELREQUEST,
    output_type=message__pb2._RESPONSE,
    serialized_options=b'\202\323\344\223\002/2*/aiflow/model-center/register-model/update:\001*',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='deleteRegisteredModel',
    full_name='ai_flow.ModelCenterService.deleteRegisteredModel',
    index=2,
    containing_service=None,
    input_type=_DELETEREGISTEREDMODELREQUEST,
    output_type=message__pb2._RESPONSE,
    serialized_options=b'\202\323\344\223\002/2*/aiflow/model-center/register-model/update:\001*',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='listRegisteredModels',
    full_name='ai_flow.ModelCenterService.listRegisteredModels',
    index=3,
    containing_service=None,
    input_type=_LISTREGISTEREDMODELSREQUEST,
    output_type=message__pb2._RESPONSE,
    serialized_options=b'\202\323\344\223\002*\022(/aiflow/model-center/register-model/list',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='getRegisteredModelDetail',
    full_name='ai_flow.ModelCenterService.getRegisteredModelDetail',
    index=4,
    containing_service=None,
    input_type=_GETREGISTEREDMODELDETAILREQUEST,
    output_type=message__pb2._RESPONSE,
    serialized_options=b'\202\323\344\223\002)\022\'/aiflow/model-center/register-model/get',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='createModelVersion',
    full_name='ai_flow.ModelCenterService.createModelVersion',
    index=5,
    containing_service=None,
    input_type=_CREATEMODELVERSIONREQUEST,
    output_type=message__pb2._RESPONSE,
    serialized_options=b'\202\323\344\223\002.\")/aiflow/model-center/model-version/create:\001*',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='updateModelVersion',
    full_name='ai_flow.ModelCenterService.updateModelVersion',
    index=6,
    containing_service=None,
    input_type=_UPDATEMODELVERSIONREQUEST,
    output_type=message__pb2._RESPONSE,
    serialized_options=b'\202\323\344\223\002.2)/aiflow/model-center/model-version/update:\001*',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='deleteModelVersion',
    full_name='ai_flow.ModelCenterService.deleteModelVersion',
    index=7,
    containing_service=None,
    input_type=_DELETEMODELVERSIONREQUEST,
    output_type=message__pb2._RESPONSE,
    serialized_options=b'\202\323\344\223\002.2)/aiflow/model-center/model-version/delete:\001*',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='getModelVersionDetail',
    full_name='ai_flow.ModelCenterService.getModelVersionDetail',
    index=8,
    containing_service=None,
    input_type=_GETMODELVERSIONDETAILREQUEST,
    output_type=message__pb2._RESPONSE,
    serialized_options=b'\202\323\344\223\002(\022&/aiflow/model-center/model-version/get',
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_MODELCENTERSERVICE)

DESCRIPTOR.services_by_name['ModelCenterService'] = _MODELCENTERSERVICE

ModelCenterService = service_reflection.GeneratedServiceType('ModelCenterService', (_service.Service,), dict(
  DESCRIPTOR = _MODELCENTERSERVICE,
  __module__ = 'model_center_service_pb2'
  ))

ModelCenterService_Stub = service_reflection.GeneratedServiceStubType('ModelCenterService_Stub', (ModelCenterService,), dict(
  DESCRIPTOR = _MODELCENTERSERVICE,
  __module__ = 'model_center_service_pb2'
  ))


# @@protoc_insertion_point(module_scope)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from typing import Text
import traceback

from ai_flow.common.configuration import AIFlowConfiguration
from ai_flow.workflow.workflow import WorkflowPropertyKeys
from ai_flow.plugin_interface.blob_manager_interface import BlobConfig, BlobManagerFactory

from ai_flow.util import json_utils
from ai_flow.context.project_context import ProjectContext, build_project_context
from ai_flow.protobuf.message_pb2 import ResultProto, StatusProto
from ai_flow.protobuf.scheduling_service_pb2_grpc import SchedulingServiceServicer
from ai_flow.protobuf.scheduling_service_pb2 import \
    (ScheduleWorkflowRequest,
     WorkflowInfoResponse,
     ListWorkflowInfoResponse,
     WorkflowExecutionRequest,
     WorkflowExecutionResponse,
     ListWorkflowExecutionResponse,
     ScheduleJobRequest,
     JobInfoResponse,
     ListJobInfoResponse)
from ai_flow.scheduler.scheduler_factory import SchedulerFactory
from ai_flow.plugin_interface.scheduler_interface import Scheduler
from ai_flow.workflow.workflow import Workflow
from ai_flow.endpoint.server.workflow_proto_utils import workflow_to_proto, workflow_list_to_proto, \
    workflow_execution_to_proto, workflow_execution_list_to_proto, job_to_proto, job_list_to_proto


class SchedulerServiceConfig(AIFlowConfiguration):

    def __init__(self):
        super().__init__()
        self['scheduler_config'] = {}

    def repository(self):
        if 'repository' not in self:
            return '/tmp'
        else:
            return self['repository']

    def set_repository(self, value):
        self['repository'] = value

    def scheduler_class(self):
        if self.get('scheduler_class') is not None:
            return self.get('scheduler_class')
        else:
            return None

    def set_scheduler_class(self, value):
        self['scheduler_class'] = value

    def scheduler_config(self):
        if 'scheduler_config' not in self:
            return None
        return self['scheduler_config']

    def set_scheduler_config(self, value):
        self['scheduler_config'] = value


class SchedulerService(SchedulingServiceServicer):
    def __init__(self,
                 scheduler_service_config: SchedulerServiceConfig):
        self._scheduler_service_config = scheduler_service_config
        self._scheduler: Scheduler \
            = SchedulerFactory.create_scheduler(scheduler_service_config.scheduler_class(),
                                                scheduler_service_config.scheduler_config())

    # workflow interface
    def submitWorkflow(self, request, context):
        try:
            rq: ScheduleWorkflowRequest = request
            if rq.workflow_json is None or '' == rq.workflow_json:
                return WorkflowInfoResponse(result=ResultProto(status=StatusProto.ERROR,
                                                               error_message='workflow json is empty!'))
            workflow: Workflow = json_utils.loads(rq.workflow_json)
            raw_config = {}
            raw_config.update(workflow.properties[WorkflowPropertyKeys.BLOB])
            blob_config = BlobConfig(raw_config)
            blob_manager = BlobManagerFactory.create_blob_manager(blob_config.blob_manager_class(),
                                                                  blob_config.blob_manager_config())
            project_path: Text = blob_manager\
                .download_project(workflow_snapshot_id=workflow.workflow_snapshot_id,
                                  remote_path=workflow.project_uri,
                                  local_path=self._scheduler_service_config.repository())

            project_context: ProjectContext = build_project_context(project_path)
            project_name = project_context.project_name

            workflow_info = self._scheduler.submit_workflow(workflow, project_context)
            if workflow_info is None:
                return WorkflowInfoResponse(
                    result=ResultProto(status=StatusProto.ERROR,
                                       error_message='{}, {} do not exist!'.format(project_name,
                                                                                   workflow.workflow_name)))
            return WorkflowInfoResponse(result=ResultProto(status=StatusProto.OK),
                                        workflow=workflow_to_proto(workflow_info))
        except Exception as err:
            return WorkflowInfoResponse(result=ResultProto(status=StatusProto.ERROR,
                                                           error_message=traceback.format_exc()))

    def deleteWorkflow(self, request, context):
        try:
            rq: ScheduleWorkflowRequest = request
            workflow_info = self._scheduler.delete_workflow(rq.namespace, rq.workflow_name)
            if workflow_info is None:
                return WorkflowInfoResponse(
                    result=ResultProto(status=StatusProto.ERROR,
                                       error_message='{} do not exist!'.format(rq.workflow_name)))
            return WorkflowInfoResponse(result=ResultProto(status=StatusProto.OK),
                                        workflow=workflow_to_proto(workflow_info))
        except Exception as err:
            return WorkflowInfoResponse(result=ResultProto(status=StatusProto.ERROR,
                                                           error_message=traceback.format_exc()))

    def pauseWorkflowScheduling(self, request, context):
        try:
            rq: ScheduleWorkflowRequest = request
            workflow_info = self._scheduler.pause_workflow_scheduling(rq.namespace, rq.workflow_name)
            if workflow_info is None:
                return WorkflowInfoResponse(
                    result=ResultProto(status=StatusProto.ERROR,
                                       error_message='{} do not exist!'.format(rq.workflow_name)))
            return WorkflowInfoResponse(result=ResultProto(status=StatusProto.OK),
                                        workflow=workflow_to_proto(workflow_info))
        except Exception as err:
            return WorkflowInfoResponse(result=ResultProto(status=StatusProto.ERROR,
                                                           error_message=traceback.format_exc()))

    def resumeWorkflowScheduling(self, request, context):
        try:
            rq: ScheduleWorkflowRequest = request
            workflow_info = self._scheduler.resume_workflow_scheduling(rq.namespace, rq.workflow_name)
            if workflow_info is None:
                return WorkflowInfoResponse(
                    result=ResultProto(status=StatusProto.ERROR,
                                       error_message='{} do not exist!'.format(rq.workflow_name)))
            return WorkflowInfoResponse(result=ResultProto(status=StatusProto.OK),
                                        workflow=workflow_to_proto(workflow_info))
        except Exception as err:
            return WorkflowInfoResponse(result=ResultProto(status=StatusProto.ERROR,
                                                           error_message=traceback.format_exc()))

    def getWorkflow(self, request, context):
        return WorkflowInfoResponse(result=ResultProto(status=StatusProto.ERROR,
                                                       error_message='Do not support getWorkflow'))

    def listWorkflows(self, request, context):
        return WorkflowInfoResponse(result=ResultProto(status=StatusProto.ERROR,
                                                       error_message='Do not support listWorkflows'))

    # workflow execution interface

    def startNewWorkflowExecution(self, request, context):
        try:
            rq: WorkflowExecutionRequest = request
            workflow_execution = self._scheduler.start_new_workflow_execution(rq.namespace, rq.workflow_name)
            if workflow_execution is None:
                return WorkflowInfoResponse(
                    result=ResultProto(status=StatusProto.ERROR,
                                       error_message='{}, {} do not exist!'.format(rq.namespace, rq.workflow_name)))
            return WorkflowExecutionResponse(result=ResultProto(status=StatusProto.OK),
                                             workflow_execution=workflow_execution_to_proto(workflow_execution))
        except Exception as err:
            return WorkflowExecutionResponse(result=ResultProto(status=StatusProto.ERROR,
                                                                error_message=traceback.format_exc()))

    def killAllWorkflowExecutions(self, request, context):
        try:
            rq: WorkflowExecutionRequest = request
            workflow_execution_list = self._scheduler.stop_all_workflow_execution(rq.namespace, rq.workflow_name)
            response = ListWorkflowExecutionResponse(result=ResultProto(status=StatusProto.OK))
            response.workflow_execution_list.extend(workflow_execution_list_to_proto(workflow_execution_list))
            return response
        except Exception as err:
            return ListWorkflowExecutionResponse(result=ResultProto(status=StatusProto.ERROR,
                                                                    error_message=traceback.format_exc()))

    def killWorkflowExecution(self, request, context):
        try:
            rq: WorkflowExecutionRequest = request
            workflow_execution = self._scheduler.stop_workflow_execution(rq.execution_id)
            if workflow_execution is None:
                return WorkflowExecutionResponse(
                    result=ResultProto(status=StatusProto.ERROR,
                                       error_message='{} do not exist!'.format(rq.execution_id)))
            return WorkflowExecutionResponse(result=ResultProto(status=StatusProto.OK),
                                             workflow_execution=workflow_execution_to_proto(workflow_execution))
        except Exception as err:
            return WorkflowExecutionResponse(result=ResultProto(status=StatusProto.ERROR,
                                                                error_message=traceback.format_exc()))

    def getWorkflowExecution(self, request, context):
        try:
            rq: WorkflowExecutionRequest = request
            workflow_execution = self._scheduler.get_workflow_execution(rq.execution_id)
            if workflow_execution is None:
                return WorkflowExecutionResponse(
                    result=ResultProto(status=StatusProto.ERROR,
                                       error_message='{} do not exist!'.format(rq.execution_id)))
            return WorkflowExecutionResponse(result=ResultProto(status=StatusProto.OK),
                                             workflow_execution=workflow_execution_to_proto(workflow_execution))
        except Exception as err:
            return WorkflowExecutionResponse(result=ResultProto(status=StatusProto.ERROR,
                                                                error_message=traceback.format_exc()))

    def listWorkflowExecutions(self, request, context):
        try:
            rq: WorkflowExecutionRequest = request
            workflow_execution_list = self._scheduler.list_workflow_executions(rq.namespace, rq.workflow_name)
            response = ListWorkflowExecutionResponse(result=ResultProto(status=StatusProto.OK))
            response.workflow_execution_list.extend(workflow_execution_list_to_proto(workflow_execution_list))
            return response
        except Exception as err:
            return ListWorkflowExecutionResponse(result=ResultProto(status=StatusProto.ERROR,
                                                                    error_message=traceback.format_exc()))

    # job interface
    def startJob(self, request, context):
        try:
            rq: ScheduleJobRequest = request
            job = self._scheduler.start_job_execution(rq.job_name, rq.execution_id)
            if job is None:
                return JobInfoResponse(
                    result=ResultProto(status=StatusProto.ERROR,
                                       error_message='{} do not exist!'.format(rq.job_name)))
            return JobInfoResponse(result=ResultProto(status=StatusProto.OK), job=job_to_proto(job))
        except Exception as err:
            return JobInfoResponse(result=ResultProto(status=StatusProto.ERROR,
                                                      error_message=traceback.format_exc()))

    def stopJob(self, request, context):
        try:
            rq: ScheduleJobRequest = request
            job = self._scheduler.stop_job_execution(rq.job_name, rq.execution_id)
            if job is None:
                return JobInfoResponse(
                    result=ResultProto(status=StatusProto.ERROR,
                                       error_message='{} do not exist!'.format(rq.job_name)))
            return JobInfoResponse(result=ResultProto(status=StatusProto.OK), job=job_to_proto(job))
        except Exception as err:
            return JobInfoResponse(result=ResultProto(status=StatusProto.ERROR,
                                                      error_message=traceback.format_exc()))

    def restartJob(self, request, context):
        try:
            rq: ScheduleJobRequest = request
            job = self._scheduler.restart_job_execution(rq.job_name, rq.execution_id)
            if job is None:
                return JobInfoResponse(
                    result=ResultProto(status=StatusProto.ERROR,
                                       error_message='{} do not exist!'.format(rq.job_name)))
            return JobInfoResponse(result=ResultProto(status=StatusProto.OK), job=job_to_proto(job))
        except Exception as err:
            return JobInfoResponse(result=ResultProto(status=StatusProto.ERROR,
                                                      error_message=traceback.format_exc()))

    def getJob(self, request, context):
        try:
            rq: ScheduleJobRequest = request
            jobs = self._scheduler.get_job_executions(rq.job_name, rq.execution_id)
            if jobs is None:
                return JobInfoResponse(
                    result=ResultProto(status=StatusProto.ERROR,
                                       error_message='{} do not exist!'.format(rq.job_name)))
            return JobInfoResponse(result=ResultProto(status=StatusProto.OK), job=job_to_proto(jobs[0]))
        except Exception as err:
            return JobInfoResponse(result=ResultProto(status=StatusProto.ERROR,
                                                      error_message=traceback.format_exc()))

    def listJobs(self, request, context):
        try:
            rq: ScheduleJobRequest = request
            job_list = self._scheduler.list_job_executions(rq.execution_id)
            response = ListJobInfoResponse(result=ResultProto(status=StatusProto.OK))
            response.job_list.extend(job_list_to_proto(job_list))
            return response
        except Exception as err:
            return ListJobInfoResponse(result=ResultProto(status=StatusProto.ERROR,
                                                          error_message=traceback.format_exc()))



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""add dagrun

Revision ID: 1b38cef5b76e
Revises: 52d714495f0
Create Date: 2015-10-27 08:31:48.475140

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = '1b38cef5b76e'
down_revision = '502898887f84'
branch_labels = None
depends_on = None


def upgrade():  # noqa: D103
    op.create_table(
        'dag_run',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('dag_id', sa.String(length=250), nullable=True),
        sa.Column('execution_date', sa.DateTime(), nullable=True),
        sa.Column('state', sa.String(length=50), nullable=True),
        sa.Column('run_id', sa.String(length=250), nullable=True),
        sa.Column('external_trigger', sa.Boolean(), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('dag_id', 'execution_date'),
        sa.UniqueConstraint('dag_id', 'run_id'),
    )


def downgrade():  # noqa: D103
    op.drop_table('dag_run')



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import logging
import socket
import string
import textwrap
import time
from functools import wraps
from typing import TYPE_CHECKING, Callable, Optional, TypeVar, cast

from airflow.configuration import conf
from airflow.exceptions import AirflowConfigException, InvalidStatsNameException
from airflow.typing_compat import Protocol

log = logging.getLogger(__name__)


class TimerProtocol(Protocol):
    """Type protocol for StatsLogger.timer"""

    def __enter__(self):
        ...

    def __exit__(self, exc_type, exc_value, traceback):
        ...

    def start(self):
        """Start the timer"""
        ...

    def stop(self, send=True):
        """Stop, and (by default) submit the timer to statsd"""
        ...


class StatsLogger(Protocol):
    """This class is only used for TypeChecking (for IDEs, mypy, pylint, etc)"""

    @classmethod
    def incr(cls, stat: str, count: int = 1, rate: int = 1) -> None:
        """Increment stat"""

    @classmethod
    def decr(cls, stat: str, count: int = 1, rate: int = 1) -> None:
        """Decrement stat"""

    @classmethod
    def gauge(cls, stat: str, value: float, rate: int = 1, delta: bool = False) -> None:
        """Gauge stat"""

    @classmethod
    def timing(cls, stat: str, dt) -> None:
        """Stats timing"""

    @classmethod
    def timer(cls, *args, **kwargs) -> TimerProtocol:
        """Timer metric that can be cancelled"""


class Timer:
    """
    Timer that records duration, and optional sends to statsd backend.

    This class lets us have an accurate timer with the logic in one place (so
    that we don't use datetime math for duration -- it is error prone).

    Example usage:

    .. code-block:: python

        with Stats.timer() as t:
            # Something to time
            frob_the_foos()

        log.info("Frobbing the foos took %.2f", t.duration)

    Or without a context manager:

    .. code-block:: python

        timer = Stats.timer().start()

        # Something to time
        frob_the_foos()

        timer.end()

        log.info("Frobbing the foos took %.2f", timer.duration)

    To send a metric:

    .. code-block:: python

        with Stats.timer("foos.frob"):
            # Something to time
            frob_the_foos()

    Or both:

    .. code-block:: python

        with Stats.timer("foos.frob") as t:
            # Something to time
            frob_the_foos()

        log.info("Frobbing the foos took %.2f", t.duration)
    """

    # pystatsd and dogstatsd both have a timer class, but present different API
    # so we can't use this as a mixin on those, instead this class is contains the "real" timer

    _start_time: Optional[int]
    duration: Optional[int]

    def __init__(self, real_timer=None):
        self.real_timer = real_timer

    def __enter__(self):
        return self.start()

    def __exit__(self, exc_type, exc_value, traceback):
        self.stop()

    def start(self):
        """Start the timer"""
        if self.real_timer:
            self.real_timer.start()
        self._start_time = time.perf_counter()
        return self

    def stop(self, send=True):  # pylint: disable=unused-argument
        """Stop the timer, and optionally send it to stats backend"""
        self.duration = time.perf_counter() - self._start_time
        if send and self.real_timer:
            self.real_timer.stop()


class DummyStatsLogger:
    """If no StatsLogger is configured, DummyStatsLogger is used as a fallback"""

    @classmethod
    def incr(cls, stat, count=1, rate=1):
        """Increment stat"""

    @classmethod
    def decr(cls, stat, count=1, rate=1):
        """Decrement stat"""

    @classmethod
    def gauge(cls, stat, value, rate=1, delta=False):
        """Gauge stat"""

    @classmethod
    def timing(cls, stat, dt):
        """Stats timing"""

    @classmethod
    def timer(cls, *args, **kwargs):
        """Timer metric that can be cancelled"""
        return Timer()


# Only characters in the character set are considered valid
# for the stat_name if stat_name_default_handler is used.
ALLOWED_CHARACTERS = set(string.ascii_letters + string.digits + '_.-')


def stat_name_default_handler(stat_name, max_length=250) -> str:
    """A function that validate the statsd stat name, apply changes to the stat name
    if necessary and return the transformed stat name.
    """
    if not isinstance(stat_name, str):
        raise InvalidStatsNameException('The stat_name has to be a string')
    if len(stat_name) > max_length:
        raise InvalidStatsNameException(
            textwrap.dedent(
                """\
            The stat_name ({stat_name}) has to be less than {max_length} characters.
        """.format(
                    stat_name=stat_name, max_length=max_length
                )
            )
        )
    if not all((c in ALLOWED_CHARACTERS) for c in stat_name):
        raise InvalidStatsNameException(
            textwrap.dedent(
                """\
            The stat name ({stat_name}) has to be composed with characters in
            {allowed_characters}.
            """.format(
                    stat_name=stat_name, allowed_characters=ALLOWED_CHARACTERS
                )
            )
        )
    return stat_name


def get_current_handler_stat_name_func() -> Callable[[str], str]:
    """Get Stat Name Handler from airflow.cfg"""
    return conf.getimport('metrics', 'stat_name_handler') or stat_name_default_handler


T = TypeVar("T", bound=Callable)  # pylint: disable=invalid-name


def validate_stat(fn: T) -> T:
    """Check if stat name contains invalid characters.
    Log and not emit stats if name is invalid
    """

    @wraps(fn)
    def wrapper(_self, stat=None, *args, **kwargs):
        try:
            if stat is not None:
                handler_stat_name_func = get_current_handler_stat_name_func()
                stat = handler_stat_name_func(stat)
            return fn(_self, stat, *args, **kwargs)
        except InvalidStatsNameException:
            log.error('Invalid stat name: %s.', stat, exc_info=True)
            return None

    return cast(T, wrapper)


class AllowListValidator:
    """Class to filter unwanted stats"""

    def __init__(self, allow_list=None):
        if allow_list:
            self.allow_list = tuple([item.strip().lower() for item in allow_list.split(',')])
        else:
            self.allow_list = None

    def test(self, stat):
        """Test if stat is in the Allow List"""
        if self.allow_list is not None:
            return stat.strip().lower().startswith(self.allow_list)
        else:
            return True  # default is all metrics allowed


class SafeStatsdLogger:
    """Statsd Logger"""

    def __init__(self, statsd_client, allow_list_validator=AllowListValidator()):
        self.statsd = statsd_client
        self.allow_list_validator = allow_list_validator

    @validate_stat
    def incr(self, stat, count=1, rate=1):
        """Increment stat"""
        if self.allow_list_validator.test(stat):
            return self.statsd.incr(stat, count, rate)
        return None

    @validate_stat
    def decr(self, stat, count=1, rate=1):
        """Decrement stat"""
        if self.allow_list_validator.test(stat):
            return self.statsd.decr(stat, count, rate)
        return None

    @validate_stat
    def gauge(self, stat, value, rate=1, delta=False):
        """Gauge stat"""
        if self.allow_list_validator.test(stat):
            return self.statsd.gauge(stat, value, rate, delta)
        return None

    @validate_stat
    def timing(self, stat, dt):
        """Stats timing"""
        if self.allow_list_validator.test(stat):
            return self.statsd.timing(stat, dt)
        return None

    @validate_stat
    def timer(self, stat=None, *args, **kwargs):
        """Timer metric that can be cancelled"""
        if stat and self.allow_list_validator.test(stat):
            return Timer(self.statsd.timer(stat, *args, **kwargs))
        return Timer()


class SafeDogStatsdLogger:
    """DogStatsd Logger"""

    def __init__(self, dogstatsd_client, allow_list_validator=AllowListValidator()):
        self.dogstatsd = dogstatsd_client
        self.allow_list_validator = allow_list_validator

    @validate_stat
    def incr(self, stat, count=1, rate=1, tags=None):
        """Increment stat"""
        if self.allow_list_validator.test(stat):
            tags = tags or []
            return self.dogstatsd.increment(metric=stat, value=count, tags=tags, sample_rate=rate)
        return None

    @validate_stat
    def decr(self, stat, count=1, rate=1, tags=None):
        """Decrement stat"""
        if self.allow_list_validator.test(stat):
            tags = tags or []
            return self.dogstatsd.decrement(metric=stat, value=count, tags=tags, sample_rate=rate)
        return None

    @validate_stat
    def gauge(self, stat, value, rate=1, delta=False, tags=None):  # pylint: disable=unused-argument
        """Gauge stat"""
        if self.allow_list_validator.test(stat):
            tags = tags or []
            return self.dogstatsd.gauge(metric=stat, value=value, tags=tags, sample_rate=rate)
        return None

    @validate_stat
    def timing(self, stat, dt, tags=None):
        """Stats timing"""
        if self.allow_list_validator.test(stat):
            tags = tags or []
            return self.dogstatsd.timing(metric=stat, value=dt, tags=tags)
        return None

    @validate_stat
    def timer(self, stat=None, *args, tags=None, **kwargs):
        """Timer metric that can be cancelled"""
        if stat and self.allow_list_validator.test(stat):
            tags = tags or []
            return Timer(self.dogstatsd.timer(stat, *args, tags=tags, **kwargs))
        return Timer()


class _Stats(type):
    instance: Optional[StatsLogger] = None

    def __getattr__(cls, name):
        return getattr(cls.instance, name)

    def __init__(cls, *args, **kwargs):
        super().__init__(cls)
        if cls.__class__.instance is None:
            try:
                is_datadog_enabled_defined = conf.has_option('metrics', 'statsd_datadog_enabled')
                if is_datadog_enabled_defined and conf.getboolean('metrics', 'statsd_datadog_enabled'):
                    cls.__class__.instance = cls.get_dogstatsd_logger()
                elif conf.getboolean('metrics', 'statsd_on'):
                    cls.__class__.instance = cls.get_statsd_logger()
                else:
                    cls.__class__.instance = DummyStatsLogger()
            except (socket.gaierror, ImportError) as e:
                log.error("Could not configure StatsClient: %s, using DummyStatsLogger instead.", e)
                cls.__class__.instance = DummyStatsLogger()

    @classmethod
    def get_statsd_logger(cls):
        """Returns logger for statsd"""
        # no need to check for the scheduler/statsd_on -> this method is only called when it is set
        # and previously it would crash with None is callable if it was called without it.
        from statsd import StatsClient

        stats_class = conf.getimport('metrics', 'statsd_custom_client_path', fallback=None)

        if stats_class:
            if not issubclass(stats_class, StatsClient):
                raise AirflowConfigException(
                    "Your custom Statsd client must extend the statsd.StatsClient in order to ensure "
                    "backwards compatibility."
                )
            else:
                log.info("Successfully loaded custom Statsd client")

        else:
            stats_class = StatsClient

        statsd = stats_class(
            host=conf.get('metrics', 'statsd_host'),
            port=conf.getint('metrics', 'statsd_port'),
            prefix=conf.get('metrics', 'statsd_prefix'),
        )
        allow_list_validator = AllowListValidator(conf.get('metrics', 'statsd_allow_list', fallback=None))
        return SafeStatsdLogger(statsd, allow_list_validator)

    @classmethod
    def get_dogstatsd_logger(cls):
        """Get DataDog statsd logger"""
        from datadog import DogStatsd

        dogstatsd = DogStatsd(
            host=conf.get('metrics', 'statsd_host'),
            port=conf.getint('metrics', 'statsd_port'),
            namespace=conf.get('metrics', 'statsd_prefix'),
            constant_tags=cls.get_constant_tags(),
        )
        dogstatsd_allow_list = conf.get('metrics', 'statsd_allow_list', fallback=None)
        allow_list_validator = AllowListValidator(dogstatsd_allow_list)
        return SafeDogStatsdLogger(dogstatsd, allow_list_validator)

    @classmethod
    def get_constant_tags(cls):
        """Get constant DataDog tags to add to all stats"""
        tags = []
        tags_in_string = conf.get('metrics', 'statsd_datadog_tags', fallback=None)
        if tags_in_string is None or tags_in_string == '':
            return tags
        else:
            for key_value in tags_in_string.split(','):
                tags.append(key_value)
            return tags


if TYPE_CHECKING:
    Stats: StatsLogger
else:

    class Stats(metaclass=_Stats):  # noqa: D101
        """Empty class for Stats - we use metaclass to inject the right one"""



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Example DAG demonstrating the usage of the BranchPythonOperator."""

import random

from airflow import DAG
from airflow.operators.dummy import DummyOperator
from airflow.operators.python import BranchPythonOperator
from airflow.utils.dates import days_ago

args = {
    'owner': 'airflow',
}

dag = DAG(
    dag_id='example_branch_operator',
    default_args=args,
    start_date=days_ago(2),
    schedule_interval="@daily",
    tags=['example', 'example2'],
)

run_this_first = DummyOperator(
    task_id='run_this_first',
    dag=dag,
)

options = ['branch_a', 'branch_b', 'branch_c', 'branch_d']

branching = BranchPythonOperator(
    task_id='branching',
    python_callable=lambda: random.choice(options),
    dag=dag,
)
run_this_first >> branching

join = DummyOperator(
    task_id='join',
    trigger_rule='none_failed_or_skipped',
    dag=dag,
)

for option in options:
    t = DummyOperator(
        task_id=option,
        dag=dag,
    )

    dummy_follow = DummyOperator(
        task_id='follow_' + option,
        dag=dag,
    )

    branching >> t >> dummy_follow >> join



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
from typing import Union, Text, Tuple, Optional, List

from notification_service.base_notification import UNDEFINED_EVENT_TYPE, ANY_CONDITION
from ai_flow.client.ai_flow_client import get_ai_flow_client
from ai_flow.ai_graph.ai_node import AINode, ReadDatasetNode, WriteDatasetNode
from ai_flow.graph.channel import Channel
from ai_flow.workflow.control_edge import ControlEdge, \
    JobAction, EventLife, ValueCondition, DEFAULT_NAMESPACE, AIFlowInternalEventType, \
    EventCondition, SchedulingRule, MeetAnyEventCondition
from ai_flow.ai_graph.ai_graph import current_graph, add_ai_node_to_graph
from ai_flow.meta.dataset_meta import DatasetMeta
from ai_flow.meta.model_meta import ModelMeta, ModelVersionMeta
from ai_flow.context.project_context import current_project_config
from ai_flow.context.workflow_config_loader import current_workflow_config
from ai_flow.workflow.status import Status


def read_dataset(dataset_info: Union[DatasetMeta, Text, int],
                 read_dataset_processor=None,
                 name: Text = None) -> Channel:
    """
    Reads the dataset through the user-defined processor based on the metadata of dataset. The processor reads dataset
    from external system and generates the :class:`~ai_flow.ai_graph.ai_node.ReadDatasetNode` with
    :class:`~ai_flow.meta.dataset_meta.DatasetMeta`.

    :param dataset_info: Information about the dataset read by the processor. The type of dataset information could be
                                      :class:`~ai_flow.meta.dataset_meta.DatasetMeta` or Text(i.e. the name of dataset)
                                      or int(i.e. the id of dataset). If the type is Text or int, this method will get
                                      the metadata of dataset by the name or id of dataset by default.
    :param read_dataset_processor: The user-defined processor composed of the logic that reads dataset.
    :param name: Name of the reading dataset operator represent for the
                            :class:`~ai_flow.ai_graph.ai_node.ReadDatasetNode`.
    :return: The :class:`~ai_flow.graph.channel.Channel`.
    """
    if isinstance(dataset_info, DatasetMeta):
        dataset_meta = dataset_info
    elif isinstance(dataset_info, Text):
        dataset_meta = get_ai_flow_client().get_dataset_by_name(dataset_info)
    else:
        dataset_meta = get_ai_flow_client().get_dataset_by_id(dataset_info)

    node = ReadDatasetNode(name=name,
                           processor=read_dataset_processor,
                           properties=None,
                           node_type='read_dataset',
                           dataset=dataset_meta)
    add_ai_node_to_graph(node, inputs=None)
    outputs = node.outputs()
    output: Channel = outputs[0]
    return output


def write_dataset(input: Channel,
                  dataset_info: Union[DatasetMeta, Text, int],
                  write_dataset_processor=None,
                  name: Text = None
                  ) -> None:
    """
    Writes the dataset through the user-defined processor based on the metadata of dataset. The processor writes dataset
    to external system and generates the :class:`~ai_flow.ai_graph.ai_node.WriteDatasetNode` with
    :class:`~ai_flow.meta.dataset_meta.DatasetMeta`.

    :param input: The input :class:`~ai_flow.graph.channel.Channel` from the operator that generates the data used to
                            write dataset.
    :param dataset_info: Information about the dataset writen by the processor. The type of dataset information could be
                                      :class:`~ai_flow.meta.dataset_meta.DatasetMeta` or Text(i.e. the name of dataset)
                                      or int(i.e. the id of dataset). If the type is Text or int, this method will get
                                      the metadata of dataset by the name or id of dataset by default.
    :param write_dataset_processor: The user-defined processor composed of the logic that writes dataset.
    :param name: Name of the writing dataset operator represent for the
                            :class:`~ai_flow.ai_graph.ai_node.WriteDatasetNode`.
    :return: None.
    """
    if isinstance(dataset_info, DatasetMeta):
        dataset_meta = dataset_info
    elif isinstance(dataset_info, Text):
        dataset_meta = get_ai_flow_client().get_dataset_by_name(dataset_info)
    else:
        dataset_meta = get_ai_flow_client().get_dataset_by_id(dataset_info)

    node = WriteDatasetNode(name=name,
                            processor=write_dataset_processor,
                            properties=None,
                            node_type='write_dataset',
                            dataset=dataset_meta)
    add_ai_node_to_graph(node, inputs=input)
    return None


def transform(input: Union[Channel, List[Channel]],
              transform_processor,
              output_num=1,
              name: Text = None) -> Union[Channel, Tuple[Channel]]:
    """
    Transforms dataset through the user-defined processor based on the input :class:`~ai_flow.graph.channel.Channel`.
    The processor generates the transforming :class:`~ai_flow.ai_graph.ai_node.AINode`.

    :param input: List of the input :class:`~ai_flow.graph.channel.Channel` that contains the multiple channels from the
                           operators which generate the data used to transform.
    :param transform_processor: The user-defined processor composed of the transformation logic.
    :param output_num: The output :class:`~ai_flow.graph.channel.Channel` number of the transforming operator.
                                      The default value is 1.
    :param name: Name of the transforming operator represent for the :class:`~ai_flow.ai_graph.ai_node.AINode`.
    :return: Returns the :class:`~ai_flow.graph.channel.Channel` when the `output_num` is 1 and returns
                 Tuple[:class:`~ai_flow.graph.channel.Channel`] when the `output_num` is bigger than 1.
    """
    return user_define_operation(input=input,
                                 name=name,
                                 processor=transform_processor,
                                 output_num=output_num,
                                 operation_type='transform')


def train(input: Union[Channel, List[Channel]],
          model_info: Union[ModelMeta, Text, int],
          training_processor,
          base_model_info: Union[ModelMeta, Text, int] = None,
          output_num=0,
          name: Text = None) -> Union[Channel, Tuple[Channel]]:
    """
    Trains model through the user-defined processor with the information of models based on the input
    :class:`~ai_flow.graph.channel.Channel`. The processor generates the model training
    :class:`~ai_flow.ai_graph.ai_node.AINode`.

    :param input: List of the input :class:`~ai_flow.graph.channel.Channel` that contains the multiple channels from the
                           operators which generate the data used to train model.
    :param model_info: Information about the model trained by the processor. The type of model could be
                                     :class:`~ai_flow.meta.model_meta.ModelMeta` or Text(i.e. the name of model) or int(
                                     i.e. the id of model). If the type is Text or int, this method will get the
                                     metadata of model by the name or id of model by default.
    :param training_processor: The user-defined processor composed of the logic that trains model.
    :param base_model_info: Information about the base model for training by the processor. The type of model could be
                                              :class:`~ai_flow.meta.model_meta.ModelMeta` or Text or int. The
                                              `base_model_info` means the name of model when the type is Text and it
                                              means the id of model when the type is int, which gets the metadata of
                                              model by the id or name of model.
    :param output_num: The output :class:`~ai_flow.graph.channel.Channel` number of the model training operator.
                                      The default value is 0.
    :param name: Name of the model training operator represent for the :class:`~ai_flow.ai_graph.ai_node.AINode`.
    :return: Returns the :class:`~ai_flow.graph.channel.Channel` when the `output_num` is 1 and returns
                 Tuple[:class:`~ai_flow.graph.channel.Channel`] when the `output_num` is bigger than 1.
    """
    if isinstance(model_info, ModelMeta):
        output_model_meta = model_info
    elif isinstance(model_info, Text):
        output_model_meta = get_ai_flow_client().get_model_by_name(model_info)
    else:
        output_model_meta = get_ai_flow_client().get_model_by_id(model_info)

    if base_model_info is not None:
        if isinstance(base_model_info, ModelMeta):
            base_model_meta = base_model_info
        elif isinstance(base_model_info, Text):
            base_model_meta = get_ai_flow_client().get_model_by_name(base_model_info)
        else:
            base_model_meta = get_ai_flow_client().get_model_by_id(base_model_info)
    else:
        base_model_meta = None

    return user_define_operation(input=input,
                                 name=name,
                                 processor=training_processor,
                                 model_info=output_model_meta,
                                 base_model_info=base_model_meta,
                                 output_num=output_num,
                                 operation_type='train')


def predict(input: Union[Channel, List[Channel]],
            model_info: Union[ModelMeta, Text, int],
            prediction_processor,
            model_version_info: Optional[Union[ModelVersionMeta, Text]] = None,
            output_num=1,
            name: Text = None) -> Union[Channel, Tuple[Channel]]:
    """
    Predicts through the user-defined processor with the information of model and model version based on the input
    :class:`~ai_flow.graph.channel.Channel`. The processor generates the model prediction
    :class:`~ai_flow.ai_graph.ai_node.AINode`.

    :param input: List of the input :class:`~ai_flow.graph.channel.Channel` that contains the multiple channels from the
                           operators which generate the data used to predict model.
    :param model_info: Information about the model predicted by the processor. The type of model could be
                                     :class:`~ai_flow.meta.model_meta.ModelMeta` or Text(i.e. the name of model) or int(
                                     i.e. the id of model). If the type is Text or int, this method will get the
                                     metadata of model by the name or id of model by default.
    :param prediction_processor: The user-defined processor composed of the logic that predicts model.
    :param model_version_info: Information about the model version predicted by the processor. The type of model version
                                                  could be :class:`~ai_flow.meta.model_meta.ModelVersionMeta` or Text(
                                                  i.e. the `version` field of
                                                  :class:`~ai_flow.meta.model_meta.ModelVersionMeta`). If the type is
                                                  Text, this method will get the metadata of model version by the
                                                  version of model version by default.
    :param output_num: The output :class:`~ai_flow.graph.channel.Channel` number of the model prediction operator.
                                      The default value is 1.
    :param name: Name of the model prediction operator represent for the :class:`~ai_flow.ai_graph.ai_node.AINode`.
    :return: Returns the :class:`~ai_flow.graph.channel.Channel` when the `output_num` is 1 and returns
                 Tuple[:class:`~ai_flow.graph.channel.Channel`] when the `output_num` is bigger than 1.
    """
    if isinstance(model_info, ModelMeta):
        model_meta = model_info
    elif isinstance(model_info, Text):
        model_meta = get_ai_flow_client().get_model_by_name(model_info)
    else:
        model_meta = get_ai_flow_client().get_model_by_id(model_info)

    if isinstance(model_version_info, ModelVersionMeta):
        model_version_meta = model_version_info
    elif isinstance(model_version_info, Text):
        model_version_meta = get_ai_flow_client() \
            .get_model_version_by_version(model_version_info,
                                          get_ai_flow_client().get_model_by_name(model_meta.name).uuid)
    else:
        model_version_meta = None

    return user_define_operation(input=input,
                                 name=name,
                                 model_info=model_meta,
                                 processor=prediction_processor,
                                 model_version_info=model_version_meta,
                                 output_num=output_num,
                                 operation_type='predict')


def evaluate(input: Union[Channel, List[Channel]],
             model_info: Union[ModelMeta, Text, int],
             evaluation_processor,
             output_num=0,
             name: Text = None) -> Union[Channel, Tuple[Channel]]:
    """
    Evaluates model through the user-defined processor with the information of model based on the input
    :class:`~ai_flow.graph.channel.Channel`. The processor generates the model evaluating
    :class:`~ai_flow.ai_graph.ai_node.AINode`.

    :param input: List of the input :class:`~ai_flow.graph.channel.Channel` that contains the multiple channels from the
                           operators which generate the data used to evaluate model.
    :param model_info: Information about the model evaluated by the processor. The type of model could be
                                     :class:`~ai_flow.meta.model_meta.ModelMeta` or Text(i.e. the name of model) or int(
                                     i.e. the id of model). If the type is Text or int, this method will get the
                                     metadata of model by the name or id of model by default.
    :param evaluation_processor: The user-defined processor composed of the logic that evaluates model.
    :param output_num: The output :class:`~ai_flow.graph.channel.Channel` number of the model evaluating operator.
                                      The default value is 0.
    :param name: Name of the model evaluating operator represent for the :class:`~ai_flow.ai_graph.ai_node.AINode`.
    :return: Returns the :class:`~ai_flow.graph.channel.Channel` when the `output_num` is 1 and returns
                 Tuple[:class:`~ai_flow.graph.channel.Channel`] when the `output_num` is bigger than 1.
    """
    if isinstance(model_info, ModelMeta):
        model_meta = model_info
    elif isinstance(model_info, Text):
        model_meta = get_ai_flow_client().get_model_by_name(model_info)
    else:
        model_meta = get_ai_flow_client().get_model_by_id(model_info)

    return user_define_operation(input=input,
                                 name=name,
                                 model_info=model_meta,
                                 processor=evaluation_processor,
                                 output_num=output_num,
                                 operation_type='evaluate')


def dataset_validate(input: Channel,
                     dataset_validation_processor,
                     name: Text = None
                     ) -> None:
    """
    Validates dataset through the user-defined processor based on the input :class:`~ai_flow.graph.channel.Channel`.
    The processor generates the dataset validating :class:`~ai_flow.ai_graph.ai_node.AINode`.

    :param input: List of the input :class:`~ai_flow.graph.channel.Channel` that contains the multiple channels from
                           the operators which generate the data used to validate dataset.
    :param dataset_validation_processor: The user-defined processor composed of the logic that validates dataset.
    :param name: Name of the dataset validating operator represent for the :class:`~ai_flow.ai_graph.ai_node.AINode`.
    :return: Returns the :class:`~ai_flow.graph.channel.Channel` when the `output_num` is 1 and returns
                 Tuple[:class:`~ai_flow.graph.channel.Channel`] when the `output_num` is bigger than 1.
        """
    return user_define_operation(input=input,
                                 processor=dataset_validation_processor,
                                 name=name,
                                 output_num=0,
                                 operation_type='dataset_validate'
                                 )


def model_validate(input: Union[Channel, List[Channel]],
                   model_info: Union[ModelMeta, Text, int],
                   model_validation_processor,
                   model_version_info: Optional[Union[ModelVersionMeta, Text]] = None,
                   base_model_version_info: Optional[Union[ModelVersionMeta, Text]] = None,
                   output_num=0,
                   name: Text = None) -> Union[Channel, Tuple[Channel]]:
    """
    Validates model through the user-defined processor with the information of model and model versions based on the input
    :class:`~ai_flow.graph.channel.Channel`. The processor generates the model validating
    :class:`~ai_flow.ai_graph.ai_node.AINode`.

    :param input: List of the input :class:`~ai_flow.graph.channel.Channel` that contains the multiple channels from the
                           operators which generate the data used to validate model.
    :param model_info: Information about the model validated by the processor. The type of model could be
                                     :class:`~ai_flow.meta.model_meta.ModelMeta` or Text(i.e. the name of model) or int(
                                     i.e. the id of model). If the type is Text or int, this method will get the
                                     metadata of model by the name or id of model by default.
    :param model_validation_processor: The user-defined processor composed of the logic that validates model.
    :param model_version_info: Information about the model version validated by the processor. The type of model version
                                                  could be :class:`~ai_flow.meta.model_meta.ModelVersionMeta` or Text(
                                                  i.e. the `version` field of
                                                  :class:`~ai_flow.meta.model_meta.ModelVersionMeta`). If the type is
                                                  Text, this method will get the metadata of model version by the
                                                  version of model version by default.
    :param base_model_version_info: Information about the base model version for validating by the processor. The type
                                                            of model version could be
                                                            :class:`~ai_flow.meta.model_meta.ModelVersionMeta` or Text(
                                                            i.e. the `version` field of 
                                                            :class:`~ai_flow.meta.model_meta.ModelVersionMeta`). If the
                                                            type is Text, this method will get the metadata of model
                                                            version by the version of model version by default.
    :param output_num: The output :class:`~ai_flow.graph.channel.Channel` number of the model validating operator.
                                      The default value is 0.
    :param name: Name of the model validating operator represent for the :class:`~ai_flow.ai_graph.ai_node.AINode`.
    :return: Returns the :class:`~ai_flow.graph.channel.Channel` when the `output_num` is 1 and returns
                 Tuple[:class:`~ai_flow.graph.channel.Channel`] when the `output_num` is bigger than 1.
    """
    if isinstance(model_info, ModelMeta):
        model_meta = model_info
    elif isinstance(model_info, Text):
        model_meta = get_ai_flow_client().get_model_by_name(model_info)
    else:
        model_meta = get_ai_flow_client().get_model_by_id(model_info)

    if isinstance(model_version_info, ModelVersionMeta):
        model_version_meta = model_version_info
    elif isinstance(model_version_info, Text):
        model_version_meta = get_ai_flow_client() \
            .get_model_version_by_version(model_version_info,
                                          get_ai_flow_client().get_model_by_name(model_meta.name).uuid)
    else:
        model_version_meta = None

    if isinstance(base_model_version_info, ModelVersionMeta):
        base_model_version_meta = model_version_info
    elif isinstance(base_model_version_info, Text):
        base_model_version_meta = get_ai_flow_client() \
            .get_model_version_by_version(model_version_info,
                                          get_ai_flow_client().get_model_by_name(model_meta.name).uuid)
    else:
        base_model_version_meta = None

    return user_define_operation(input=input,
                                 model_info=model_meta,
                                 processor=model_validation_processor,
                                 name=name,
                                 model_version_info=model_version_meta,
                                 base_model_version_info=base_model_version_meta,
                                 output_num=output_num,
                                 operation_type='model_validate'
                                 )


def push_model(model_info: Union[ModelMeta, Text, int],
               pushing_model_processor,
               model_version_info: Optional[Union[ModelVersionMeta, Text]] = None,
               name: Text = None) -> None:
    """
    Pushes model through the user-defined processor with the information of model version based on the input
    :class:`~ai_flow.graph.channel.Channel`. The processor generates the model pushing
    :class:`~ai_flow.ai_graph.ai_node.AINode`.

    :param model_info: Information about the model pushed by the processor. The type of model could be
                                     :class:`~ai_flow.meta.model_meta.ModelMeta` or Text(i.e. the name of model) or int(
                                     i.e. the id of model). If the type is Text or int, this method will get the
                                     metadata of model by the name or id of model by default.
    :param pushing_model_processor: The user-defined processor composed of the logic that pushes model.
    :param model_version_info: Information about the model version pushed by the processor. The type of model version
                                                  could be :class:`~ai_flow.meta.model_meta.ModelVersionMeta` or Text(
                                                  i.e. the `version` field of
                                                  :class:`~ai_flow.meta.model_meta.ModelVersionMeta`). If the type is
                                                  Text, this method will get the metadata of model version by the
                                                  version of model version by default.
    :param name: Name of the model pushing operator represent for the :class:`~ai_flow.ai_graph.ai_node.AINode`.
    :return: None.
    """
    if isinstance(model_info, ModelMeta) or model_info is None:
        model_meta = model_info
    elif isinstance(model_info, Text):
        model_meta = get_ai_flow_client().get_model_by_name(model_info)
    else:
        model_meta = get_ai_flow_client().get_model_by_id(model_info)

    if isinstance(model_version_info, ModelVersionMeta) or model_version_info is None:
        model_version_info = None
    elif isinstance(model_version_info, Text):
        model_version_info = get_ai_flow_client().get_model_version_by_version(version=model_version_info,
                                                                               model_id=model_meta.uuid)

    return user_define_operation(model_info=model_meta,
                                 processor=pushing_model_processor,
                                 model_version_info=model_version_info,
                                 name=name,
                                 operation_type='push_model')


def user_define_operation(input: Union[None, Channel, List[Channel]] = None,
                          processor=None,
                          output_num=1,
                          name: Text = None,
                          operation_type: Text = 'user_define_operation',
                          **kwargs) -> Union[None, Channel, Tuple[Channel]]:
    """
    Operation through the user-defined processor based on the input :class:`~ai_flow.graph.channel.Channel`. The
    processor generates the :class:`~ai_flow.ai_graph.ai_node.AINode`.

    :param input: List of the input :class:`~ai_flow.graph.channel.Channel` that contains the multiple channels from the
                           operators which generate the data.
    :param processor: The user-defined processor composed of the processing logic.
    :param output_num: The output :class:`~ai_flow.graph.channel.Channel` number. The default value is 1.
    :param name: Name of the operator represent for the :class:`~ai_flow.ai_graph.ai_node.AINode`.
    :param operation_type: The type of the operation.
    :param kwargs: The arguments of the operation used.
    :return: Returns None when no output and returns the :class:`~ai_flow.graph.channel.Channel` when the `output_num`
                 is 1 and returns Tuple[:class:`~ai_flow.graph.channel.Channel`] when the `output_num` is bigger than 1.
    """
    node = AINode(name=name,
                  processor=processor,
                  properties=None,
                  output_num=output_num,
                  node_type=operation_type,
                  **kwargs)
    add_ai_node_to_graph(node, inputs=input)
    outputs = node.outputs()
    if 0 == output_num:
        return None
    else:
        if 1 == len(outputs):
            output: Channel = outputs[0]
            return output
        else:
            output: List[Channel] = []
            for i in outputs:
                tmp: Channel = i
                output.append(tmp)
            return tuple(output)


def action_on_event(job_name: Text,
                    event_key: Text,
                    event_value: Text,
                    event_type: Text = UNDEFINED_EVENT_TYPE,
                    sender: Text = None,
                    namespace: Text = DEFAULT_NAMESPACE,
                    value_condition: ValueCondition = ValueCondition.EQUALS,
                    action: JobAction = JobAction.START):
    """
    When the specified event is met, the :class:`~ai_flow.workflow.control_edge.JobAction` is triggered on the job with
    the given job name. It is a convenient method to add a :class:`~ai_flow.workflow.control_edge.MeetAnyEventCondition`
    with one event to a job.

    :param job_name: The name of the job triggered.
    :param event_key: The key of the event met.
    :param event_value: The value of the event met.
    :param event_type: The type of the event met.
    :param sender: The sender of event which uses the name of the job. If sender is None, the sender is dependency.
    :param namespace: The namespace of the event which uses the name of the project.
    :param value_condition: The event :class:`~ai_flow.workflow.control_edge.ValueCondition`, whose value is EQUALS or UPDATE.
                                            EQUALS means the source channel will start or restart only in the condition
                                            that the notification service updates a value which equals to the event
                                            value under the specified event key, while UPDATE means the source channel
                                            will start or restart when in the the condition that the notification
                                            service has a update operation on the event key which event value belongs.
    :param action: The :class:`~ai_flow.workflow.control_edge.JobAction` acts on the source channel whose value includes
                             START, RESTART, STOP and NONE.
    """
    event_condition = MeetAnyEventCondition()
    event_condition.add_event(event_key, event_value, event_type, namespace, sender, EventLife.ONCE, value_condition)
    action_on_events(job_name, event_condition, action)


def action_on_events(job_name: Text, event_condition: EventCondition, action: JobAction):
    """
    Defines a rule which is the combination of a :class:`~ai_flow.workflow.control_edge.EventCondition` and
    :class:`~ai_flow.workflow.control_edge.JobAction`, on the job with the given name. When the
    :class:`~ai_flow.workflow.control_edge.EventCondition` of the rule met, the corresponding
    :class:`~ai_flow.workflow.control_edge.JobAction` will be triggered on the job. User could call the method
    multiple times on the same job to add multiples rules, the rules will be checked according to the order of the
    the method call.

    :param job_name: The name of the job to add the rule.
    :param event_condition: The :class:`~ai_flow.workflow.control_edge.EventCondition` of the rule.
    :param action: The :class:`~ai_flow.workflow.control_edge.JobAction` to take when the
                             :class:`~ai_flow.workflow.control_edge.EventCondition` is met.
    """
    rule = SchedulingRule(event_condition, action)
    control_edge = ControlEdge(destination=job_name, scheduling_rule=rule)
    current_graph().add_edge(job_name, control_edge)


def action_on_model_version_event(job_name: Text,
                                  model_name: Text,
                                  model_version_event_type: Text,
                                  namespace: Text = DEFAULT_NAMESPACE,
                                  action: JobAction = JobAction.RESTART):
    """
    Adds the model version control dependency. It means the job will take given action when the type of model version
    event is updated to the specified  type.

    :param job_name: The name of the job triggered.
    :param model_name: The name of the model refer to a specified model.
    :param model_version_event_type: The type of the model version event, whose value is of the
                                                             :class:`~ai_flow.model_center.entity.model_version_stage.ModelVersionEventType`.
    :param namespace: The namespace of the event which uses the name of the project.
    :param action: The :class:`~ai_flow.workflow.control_edge.JobAction` acts on the source channel whose value includes
                             START, RESTART, STOP and NONE.
    """
    action_on_event(job_name=job_name,
                    event_key=model_name,
                    event_value="*",
                    event_type=model_version_event_type,
                    action=action,
                    value_condition=ValueCondition.UPDATED,
                    namespace=namespace,
                    sender=ANY_CONDITION)


def action_on_dataset_event(job_name: Text,
                            dataset_name: Text,
                            namespace: Text = DEFAULT_NAMESPACE,
                            action: JobAction = JobAction.START):
    """
    Adds the dataset control dependency. It means the job will take given action when only a new dataset of the
    specified dataset is updated in notification service.

    :param job_name: The name of the job triggered.
    :param dataset_name: The name of the dataset refer to a specified dataset.
    :param namespace: The namespace of the event which uses the name of the project.
    :param action: The :class:`~ai_flow.workflow.control_edge.JobAction` acts on the source channel whose value includes
                             START, RESTART, STOP and NONE.
    """
    action_on_event(job_name=job_name,
                    event_key=dataset_name,
                    event_value="created",
                    event_type=AIFlowInternalEventType.DATASET_CHANGED,
                    action=action,
                    namespace=namespace,
                    sender=ANY_CONDITION
                    )


def action_on_job_status(job_name: Text,
                         upstream_job_name: Text,
                         upstream_job_status: Status = Status.FINISHED,
                         action: JobAction = JobAction.START):
    """
    Triggers the job with the given name when the status of the upstream job changes.

    :param job_name: The name of the job triggered.
    :param upstream_job_name: The name of the upstream job.
    :param upstream_job_status: The status of the upstream job.
    :param action: The :class:`~ai_flow.workflow.control_edge.JobAction` acts on the source channel whose value includes
                             START, RESTART, STOP and NONE.
    """
    event_key = '.'.join([current_workflow_config().workflow_name, upstream_job_name])
    action_on_event(job_name=job_name,
                    event_key=event_key,
                    event_type=AIFlowInternalEventType.JOB_STATUS_CHANGED,
                    sender=upstream_job_name,
                    event_value=upstream_job_status,
                    action=action,
                    namespace=current_project_config().get_project_name())



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
"""Roles sub-commands"""

from airflow.cli.simple_table import AirflowConsole
from airflow.utils import cli as cli_utils
from airflow.utils.cli import suppress_logs_and_warning
from airflow.www.app import cached_app


@suppress_logs_and_warning()
def roles_list(args):
    """Lists all existing roles"""
    appbuilder = cached_app().appbuilder  # pylint: disable=no-member
    roles = appbuilder.sm.get_all_roles()
    AirflowConsole().print_as(
        data=sorted([r.name for r in roles]), output=args.output, mapper=lambda x: {"name": x}
    )


@cli_utils.action_logging
@suppress_logs_and_warning()
def roles_create(args):
    """Creates new empty role in DB"""
    appbuilder = cached_app().appbuilder  # pylint: disable=no-member
    for role_name in args.role:
        appbuilder.sm.add_role(role_name)
    print(f"Added {len(args.role)} role(s)")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import threading
import time
from datetime import timedelta
from typing import List, Optional, Set

import airflow.utils.dag_processing as dag_processing
from airflow.configuration import conf
from airflow.contrib.jobs.background_service import BackgroundService
from airflow.events.scheduler_events import DagExecutableEvent
from airflow.jobs import scheduler_job
from airflow.models import DagModel
from airflow.utils.log.logging_mixin import LoggingMixin
from airflow.utils.mailbox import Mailbox
from airflow.utils.mixins import MultiprocessingStartMethodMixin
from airflow.utils.session import create_session


class DagTriggerDagFileProcessorAgent(dag_processing.DagFileProcessorAgent):
    def wait_on_manager_message(self, timeout=None):
        self._parent_signal_conn.poll(timeout)


class StoppableThread(threading.Thread, LoggingMixin):
    """Thread class with a stop() method. The thread itself has to check
    regularly for the stopped() condition."""

    def __init__(self, *args, **kwargs):
        super(StoppableThread, self).__init__(*args, **kwargs)
        self._ended = threading.Event()

    def stop(self):
        self.log.debug("stopping thread")
        self._ended.set()

    def stopped(self):
        return self._ended.is_set()


class DagRunnableReportingThread(StoppableThread, LoggingMixin):

    def __init__(self, async_mode: bool, dag_file_processor_agent, mailbox: Mailbox, *args, **kwargs):
        super(DagRunnableReportingThread, self).__init__(*args, **kwargs)
        self.setName("DagTrigger-DagRunnableReporter")
        self._async_mode = async_mode
        self._dag_file_processor_agent: DagTriggerDagFileProcessorAgent = dag_file_processor_agent
        self._mailbox = mailbox

    def run(self) -> None:
        while not self.stopped():
            # send AGENT_RUN_ONCE to DagFileProcessorManager to trigger dag parsing if not async mode
            if not self._async_mode:
                self._dag_file_processor_agent.run_single_parsing_loop()
            with create_session() as session:
                dag_models = DagModel.dags_needing_dagruns(session).all()
                self.log.debug("dags needs dagruns: {}".format(dag_models))
                self._send_dag_executable(dag_models)
            time.sleep(5)
        self.log.info("DagRunnableReporter exiting")

    def _send_dag_executable(self, dag_models: Set[DagModel]):
        for dag_model in dag_models:
            self._mailbox.send_message(DagExecutableEvent(dag_model.dag_id).to_event())


class ParsingStatRetrieveThread(StoppableThread):
    def __init__(self, dag_file_processor_agent, *args, **kwargs):
        super(ParsingStatRetrieveThread, self).__init__(*args, **kwargs)
        self.setName("DagTrigger-ParsingStatRetriever")
        self._dag_file_processor_agent: DagTriggerDagFileProcessorAgent = dag_file_processor_agent

    def run(self) -> None:
        while not self.stopped():
            self._dag_file_processor_agent.wait_on_manager_message()
            self._dag_file_processor_agent.heartbeat()
            time.sleep(10)
        self.log.info("ParsingStatRetriever exiting")


class DagTrigger(BackgroundService, MultiprocessingStartMethodMixin):

    def __init__(self,
                 dag_directory: str,
                 max_runs: int,
                 dag_ids: Optional[List[str]],
                 pickle_dags: bool,
                 mailbox: Mailbox,
                 refresh_dag_dir_interval=1,
                 notification_service_uri=None):
        """
        :param dag_directory: Directory where DAG definitions are kept. All
        files in file_paths should be under this directory
        :type dag_directory: unicode
        :param max_runs: The number of times to parse and schedule each file. -1
            for unlimited.
        :type max_runs: int
        :param dag_ids: if specified, only schedule tasks with these DAG IDs
        :type dag_ids: list[str]
        :param pickle_dags: whether to pickle DAGs.
        :type pickle_dags: bool
        :param mailbox: the mailbox to send the DagExecutableEvent
        :type mailbox: Mailbox
        """

        super().__init__()

        self._dag_directory = dag_directory
        self._max_runs = max_runs
        self._dag_ids = dag_ids
        self._pickle_dags = pickle_dags
        self._mailbox = mailbox
        # use synchronize mode when using sqlite
        self._async_mode = not conf.get('core', 'sql_alchemy_conn').lower().startswith('sqlite')

        self._dag_runnable_reporting_thread: Optional[StoppableThread] = None
        self._parsing_stat_process_thread: Optional[StoppableThread] = None
        self._dag_file_processor_agent: Optional[DagTriggerDagFileProcessorAgent] = None
        self._refresh_dag_dir_interval = refresh_dag_dir_interval
        self._notification_service_uri = notification_service_uri

    def start(self):
        self._start_dag_file_processor_manager()
        self._dag_runnable_reporting_thread = DagRunnableReportingThread(self._async_mode,
                                                                         self._dag_file_processor_agent,
                                                                         self._mailbox)
        self._dag_runnable_reporting_thread.start()

        self._parsing_stat_process_thread = ParsingStatRetrieveThread(self._dag_file_processor_agent)
        self._parsing_stat_process_thread.start()

    def end(self) -> None:
        if self._dag_file_processor_agent is not None:
            self._dag_file_processor_agent.terminate()
        if self._dag_runnable_reporting_thread is not None:
            self._dag_runnable_reporting_thread.stop()
        if self._parsing_stat_process_thread is not None:
            self._parsing_stat_process_thread.stop()
        self._dag_runnable_reporting_thread.join()
        self._parsing_stat_process_thread.join()
        self._dag_file_processor_agent.end()

    def terminate(self):
        if self._dag_file_processor_agent is not None:
            self._dag_file_processor_agent.end()
        if self._dag_runnable_reporting_thread is not None:
            self._dag_runnable_reporting_thread.stop()
        if self._parsing_stat_process_thread is not None:
            self._parsing_stat_process_thread.stop()

    def _start_dag_file_processor_manager(self):
        processor_factory = scheduler_job.SchedulerJob._create_dag_file_processor

        self._dag_file_processor_agent = DagTriggerDagFileProcessorAgent(self._dag_directory,
                                                                         self._max_runs,
                                                                         processor_factory,
                                                                         self._get_processor_timeout(),
                                                                         [],
                                                                         self._pickle_dags,
                                                                         self._async_mode,
                                                                         self._refresh_dag_dir_interval,
                                                                         self._notification_service_uri)
        self._dag_file_processor_agent.start()

    @staticmethod
    def _get_processor_timeout():
        processor_timeout_seconds: int = conf.getint('core', 'dag_file_processor_timeout')
        return timedelta(seconds=processor_timeout_seconds)

    def is_alive(self) -> bool:
        return self._dag_file_processor_agent is not None \
               and self._dag_runnable_reporting_thread is not None and self._dag_runnable_reporting_thread.is_alive() \
               and self._parsing_stat_process_thread is not None and self._parsing_stat_process_thread.is_alive()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

#
# Copyright (c) 2013, Michael Komitee
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""Kerberos authentication module"""
import logging
import os
from functools import wraps
from socket import getfqdn
from typing import Callable, Optional, Tuple, TypeVar, Union, cast

import kerberos
from flask import Response, _request_ctx_stack as stack, g, make_response, request  # type: ignore
from requests.auth import AuthBase
from requests_kerberos import HTTPKerberosAuth

from airflow.configuration import conf

log = logging.getLogger(__name__)

# pylint: disable=c-extension-no-member
CLIENT_AUTH: Optional[Union[Tuple[str, str], AuthBase]] = HTTPKerberosAuth(service='airflow')


class KerberosService:  # pylint: disable=too-few-public-methods
    """Class to keep information about the Kerberos Service initialized"""

    def __init__(self):
        self.service_name = None


# Stores currently initialized Kerberos Service
_KERBEROS_SERVICE = KerberosService()


def init_app(app):
    """Initializes application with kerberos"""
    hostname = app.config.get('SERVER_NAME')
    if not hostname:
        hostname = getfqdn()
    log.info("Kerberos: hostname %s", hostname)

    service = 'airflow'

    _KERBEROS_SERVICE.service_name = f"{service}@{hostname}"

    if 'KRB5_KTNAME' not in os.environ:
        os.environ['KRB5_KTNAME'] = conf.get('kerberos', 'keytab')

    try:
        log.info("Kerberos init: %s %s", service, hostname)
        principal = kerberos.getServerPrincipalDetails(service, hostname)
    except kerberos.KrbError as err:
        log.warning("Kerberos: %s", err)
    else:
        log.info("Kerberos API: server is %s", principal)


def _unauthorized():
    """
    Indicate that authorization is required
    :return:
    """
    return Response("Unauthorized", 401, {"WWW-Authenticate": "Negotiate"})


def _forbidden():
    return Response("Forbidden", 403)


def _gssapi_authenticate(token):
    state = None
    ctx = stack.top
    try:
        return_code, state = kerberos.authGSSServerInit(_KERBEROS_SERVICE.service_name)
        if return_code != kerberos.AUTH_GSS_COMPLETE:
            return None
        return_code = kerberos.authGSSServerStep(state, token)
        if return_code == kerberos.AUTH_GSS_COMPLETE:
            ctx.kerberos_token = kerberos.authGSSServerResponse(state)
            ctx.kerberos_user = kerberos.authGSSServerUserName(state)
            return return_code
        if return_code == kerberos.AUTH_GSS_CONTINUE:
            return kerberos.AUTH_GSS_CONTINUE
        return None
    except kerberos.GSSError:
        return None
    finally:
        if state:
            kerberos.authGSSServerClean(state)


T = TypeVar("T", bound=Callable)  # pylint: disable=invalid-name


def requires_authentication(function: T):
    """Decorator for functions that require authentication with Kerberos"""

    @wraps(function)
    def decorated(*args, **kwargs):
        header = request.headers.get("Authorization")
        if header:
            ctx = stack.top
            token = ''.join(header.split()[1:])
            return_code = _gssapi_authenticate(token)
            if return_code == kerberos.AUTH_GSS_COMPLETE:
                g.user = ctx.kerberos_user
                response = function(*args, **kwargs)
                response = make_response(response)
                if ctx.kerberos_token is not None:
                    response.headers['WWW-Authenticate'] = ' '.join(['negotiate', ctx.kerberos_token])

                return response
            if return_code != kerberos.AUTH_GSS_CONTINUE:
                return _forbidden()
        return _unauthorized()

    return cast(T, decorated)
